from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtWidgets import (QApplication, QGraphicsView,
                             QGraphicsPixmapItem, QGraphicsScene, QGraphicsPolygonItem,
                             QGraphicsEllipseItem, QGraphicsLineItem, QOpenGLWidget, QGraphicsTextItem)
from PyQt5.QtGui import QPainter, QPixmap, QPolygonF, QPen, QFont, QTransform
from PyQt5.QtCore import (QObject, QPointF, QPoint, QRectF,
                          QPropertyAnimation, pyqtProperty, QSequentialAnimationGroup,
                          QParallelAnimationGroup, QPauseAnimation, Qt)


EPS = 0.3
ETA = 1.8

class RPolygon(QObject):
    """A class that represents a polygon on screen (as a Qt5 object)

    :param points: list of points of the polygon
    :type points: list<(:class:`float`, :class:`float`)>
    :param line_color: color of the boundary of polygon
    :type line_color: :class:`QtGui.QColor`
    :param fill_color: color of the interior of polygon
    :type fill_color: :class:`QtGui.QColor`
    :param line_width: width of the boundary of polygon
    :type line_width: :class:`int`
    :param pos: initial position of polygon (used for animation). If none first vertex is chosen.
    :type pos: :class:`QtCore.QPointF`
    """
    def __init__(self, points, line_color, fill_color, line_width, text='', pos=None):
        self._points = [QPointF(p[0], p[1]) for p in points]
        self._pos = pos
        if self._pos is None:
            self._pos = self._points[0]
        # self._pos = QPointF(0.0, 0.0)
        self.fill_color = fill_color
        super().__init__()
        # The underlying QGraphicsPolygonItem
        self.polygon = QGraphicsPolygonItem()
        self.polygon.setPolygon(QPolygonF(self._points))
        self.polygon.setBrush(QtGui.QBrush(fill_color))
        # The underlying QGraphicsTextItem
        self._text: QGraphicsTextItem = QGraphicsTextItem(text)
        transform = QTransform.fromScale(EPS, -EPS)
        self._text.setTransformOriginPoint(self._pos)
        self._text.setTransform(transform)
        x, y = self.get_center_of_mass()
        self._text.setPos(QPointF(x - ETA, y + ETA))
        font = QFont("Times", 2)
        self._text.setFont(font)
        pen = QPen(QtGui.QPen(line_color))
        pen.setWidthF(line_width)
        self.polygon.setPen(pen)
        self._visible = 1

    def get_center_of_mass(self):
        x = 0
        y = 0
        for point in self._points:
            x += point.x()
            y += point.y()
        x /= len(self._points)
        y /= len(self._points)
        return x, y

    def x(self):
        """
        Return x position of the first vertex in polygon

        :return: x position of the first vertex in polygon
        :rtype: :class:`float`
        """
        return self._pos.x()


    def y(self):
        """
        Return y position of the first vertex in polygon

        :return: y position of the first vertex in polygon
        :rtype: :class:`float`
        """
        return self._pos.y()


    def points(self):
        """
        Return a list of polygon's vertices
        
        :return: list polygon vertices
        :rtype: list<:class:`QPointF`>
        """
        return self._points


    ####################################################
    # The following functions are for animation support
    ####################################################


    @pyqtProperty(QPointF)
    def pos(self):
        """
        Return the position of the first vertex of polygon
        (Required for animation support by Qt5)

        :return: position of the first vertex of polygon
        :rtype: :class:`QPointF`
        """
        return self._pos


    @pos.setter
    def pos(self, value):
        """
        Shift the entire polygon such that the first vertex is in
        the new coordinate
        (Required for animation support by Qt5)

        :param value: new position of the first polygon vertex
        :type value: :class:`QPointF`
        """
        delta_x = value.x() - self._pos.x()
        delta_y = value.y() - self._pos.y()
        self._points = [QPointF(p.x() + delta_x, p.y() + delta_y) for p in self._points]
        self.polygon.setPolygon(QPolygonF(self._points))
        x, y = self.get_center_of_mass()
        self._text.setPos(QPointF(x - 1.8, y + 1.8))
        self._pos = value


    @pyqtProperty(int)
    def visible(self):
        """
        Get the visibility of the polygon
        (Required for animation support by Qt5)

        :return: visibility of the polygon
        :rtype: :class:`int`
        """
        return self._visible


    @visible.setter
    def visible(self, value):
        """
        Set the visibility of the polygon
        (Required for animation support by Qt5)

        :param value: new visibility of the polygon
        :type value: :class:`int`
        """
        if value > 0:
            self.polygon.show()
        else:
            self.polygon.hide()
        self._visible = value

    @pyqtProperty(int)
    def text(self):
        """
        Get the robot's text description (number)
        (Required for animation support by Qt5)

        :return: robot text description
        :type: :class:`int`
        """
        if self._text.toPlainText() == '':
            return 0
        return int(self._text.toPlainText())

    @text.setter
    def text(self, val: int):
        """
        Set the robot's text description (number)
        (Required for animation support by Qt5)

        :param val: new robot text description
        :type val: :class:`int`
        """
        if val == 0:
            self._text.setPlainText('')
        else:
            self._text.setPlainText(str(val))
