from . import *
from ..bindings import *
from ..geometry_utils import collision_detection

def verify_paths(scene: Scene, paths: PathCollection):
    """
    Get a scene and a path, and verify that the given path collection is indeed a valid solution

    :param scene: scene
    :type scene: :class:`~discopygal.solvers.Scene`
    :param paths: paths to verify
    :type paths: :class:~discopygal.solvers.PathCollection`

    :return: True if paths are valid and if invalid, return a reason.
    :rtype: :class:`bool`, :class:`str`
    """
    if paths is None:
        return False, "Paths is None"

    # Check collision of each robot against obstacles
    lengths = []
    for robot in paths.paths:
        path = paths.paths[robot]
        lengths.append(len(path.points))
        cd = collision_detection.ObjectCollisionDetection(scene.obstacles, robot)
        for i in range(len(path.points) - 1):
            edge = Segment_2(path.points[i].location, path.points[i+1].location)
            if not cd.is_edge_valid(edge):
                return False, "Collision with obstacle"

    # Check that all paths are of same length
    if len(lengths) == 0:
        return False, "Empty path"
    if min(lengths) != max(lengths):
        return False, "Paths are not of the same length"
    path_len = max(lengths)

    # Check collision of any two robots
    robots = list(paths.paths.keys())
    for k in range(path_len-1):
        for i in range(len(robots)):
            for j in range(i+1, len(robots)):
                robot1 = robots[i]
                robot2 = robots[j]
                edge1 = Segment_2(paths.paths[robot1].points[k].location, paths.paths[robot1].points[k+1].location)
                edge2 = Segment_2(paths.paths[robot2].points[k].location, paths.paths[robot2].points[k+1].location)
                if collision_detection.collide_two_robots(robot1, edge1, robot2, edge2):
                    return False, "Collision with robots"
    
    # Check that start and end location is valid
    for robot in robots:
        start = paths.paths[robot].points[0].location
        end = paths.paths[robot].points[-1].location
        if robot.start != start or robot.end != end:
            return False, "Robots start/end goal mismatch"
    
    return True, ""


